package coral.counters;

import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import symlib.SymBool;
import coral.PC;
import coral.PartitionedCache;
import coral.util.Config;
import coral.util.Interval;
import coral.util.callers.IntervalSolverCaller;

public class CachedCounter extends PartitionedCache implements ModelCounter {
  
  interface Callback {
    Interval[] getDefaultInterval();
    EstimateResult callCounter(PC partition);
    EstimateResult callCounter(PC partition, List<Interval[]> paving);
  }
  
  private final boolean callIntervalSolverForPartitions;
  private final ModelCounter counter;
  private final IntervalSolverCaller isolver;
  private Callback callback;
  
  public EstimateResult processConstraint(Set<SymBool> conjuncts) {
    PC pc = new PC(new ArrayList<SymBool>(conjuncts)); 
    EstimateResult result;
    if (callIntervalSolverForPartitions) {
      try {
        List<Interval[]> paving = isolver.getPaving(pc,callback.getDefaultInterval());
        if (paving.size() == 0) {
          result = new EstimateResult(0,0);
        } else if (Config.mcMergeBoxes && paving.size() > 1) {
          List<Interval[]> merged = new ArrayList<Interval[]>();
          merged.add(Interval.mergeIntervals(paving));
          result = callback.callCounter(pc,merged);
        } else {
          result = callback.callCounter(pc,paving);
        }
      } catch (Exception e) {
        e.printStackTrace();
        throw new RuntimeException("Error while calling realpaver: " + e.getMessage());
      }
    } else {
      result = callback.callCounter(pc);
    }
    return result;
  }
  
  public CachedCounter(ModelCounter counter, boolean callIS) {
    super();
    this.counter = counter;
    this.callIntervalSolverForPartitions = callIS;
    isolver = coral.solvers.rand.Util.getIntervalSolver(Config.intervalSolver);
  }
  public void reset() {
    counter.reset();
  }
  
  public void reset(String seed) {
    counter.reset(seed);
  }

  @Override
  public EstimateResult count(PC pc) {
    throw new RuntimeException("not implemented");
  }

  @Override
  public EstimateResult count(PC pc, final List<Interval[]> boxes) {
    callback = new Callback() {
      public EstimateResult callCounter(PC partition) {
        return filterAndCount(partition, boxes, counter);
      }
      @Override
      public EstimateResult callCounter(PC partition, List<Interval[]> paving) {
        return filterAndCount(partition, paving, counter);
      }
      @Override
      public Interval[] getDefaultInterval() {
        return Interval.mergeIntervals(boxes);
      }
    };
    EstimateResult total = analyzeOne(pc.getConstraints());
//    System.out.println("[cached counter] total: " + total);
    
//    if (normalize) {
//      double domainVolume = 0;
//      
//      List<Interval[]> varBoxes = MCUtil.filterBoxes(boxes, pc.getSortedVars());
//      for (Interval[] box : varBoxes) {
//        domainVolume += Interval.computeVolume(box);
//      }
//      total = total.divide(domainVolume);
//      System.out.println("[cached counter] normalized total: " + total);
//    }
    
    return total;
  }

  @Override
  public EstimateResult count(PC pc, final Interval[] domains) {
    callback = new Callback() {
      public EstimateResult callCounter(PC partition) {
        Interval[] filteredIntervals = MCUtil.filterIntervals(domains, partition.getSortedVars());
        return counter.count(partition, filteredIntervals);
      }

      @Override
      public EstimateResult callCounter(PC partition, List<Interval[]> paving) {
        return filterAndCount(partition, paving, domains, counter);
      }

      @Override
      public Interval[] getDefaultInterval() {
        return domains;
      }
    };
    
    EstimateResult total = analyzeOne(pc.getConstraints());
//    System.out.println("[cached counter] total: " + total);
//    if (normalize) {
//      Interval[] varDomains = MCUtil.filterIntervals(domains, pc.getSortedVars());
//      double domainVolume = Interval.computeVolume(varDomains);
//      total = total.divide(domainVolume);
//      System.out.println("[cached counter] normalized total: " + total);
//    }
    
    return total;
  }

  @Override
  public EstimateResult count(PC pc, final List<Interval[]> boxes, final Interval[] domains) {
    callback = new Callback() {
      public EstimateResult callCounter(PC partition) {
        return filterAndCount(partition, boxes, domains, counter);
      }

      @Override
      public EstimateResult callCounter(PC partition, List<Interval[]> paving) {
        return filterAndCount(partition, paving, domains, counter);
      }

      @Override
      public Interval[] getDefaultInterval() {
        return domains;
      }
    };
    EstimateResult total = analyzeOne(pc.getConstraints());
//    System.out.println("[cached counter] total: " + total);
//    if (normalize) {
//      Interval[] varDomains = MCUtil.filterIntervals(domains, pc.getSortedVars());
//      double domainVolume = Interval.computeVolume(varDomains);
//      total = total.divide(domainVolume);
//      System.out.println("[cached counter] normalized total: " + total);
//    }
    
    return total;
  }
  
  private static EstimateResult filterAndCount(PC pc, List<Interval[]> boxes, ModelCounter counter) {
    List<Interval[]> filteredBoxes = MCUtil.filterBoxes(boxes, pc.getSortedVars());
    return counter.count(pc, filteredBoxes);
  }
  
  private static EstimateResult filterAndCount(PC pc, List<Interval[]> boxes, Interval[] domains, ModelCounter counter) {
    Interval[] filteredDomains = MCUtil.filterIntervals(domains, pc.getSortedVars());
    List<Interval[]> filteredBoxes = MCUtil.filterBoxes(boxes, pc.getSortedVars());
    return counter.count(pc, filteredBoxes, filteredDomains); 
  }
}
