package coral.counters;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import symlib.SymBool;
import symlib.SymLiteral;
import symlib.eval.Elem;
import symlib.eval.GenReversePolishExpression;
import symlib.eval.ReversePolish;
import symlib.parser.ParseException;
import symlib.parser.Parser;
import umontreal.iro.lecuyer.rng.MRG32k3a;
import umontreal.iro.lecuyer.rng.RandomStream;
import umontreal.iro.lecuyer.stat.ObservationListener;
import umontreal.iro.lecuyer.stat.StatProbe;
import umontreal.iro.lecuyer.stat.Tally;
import coral.PC;
import coral.counters.ResultData.Entry;
import coral.util.Config;
import coral.util.Interval;
import coral.util.callers.IntervalSolverCaller;

public class MonteCarloCounter implements ModelCounter {

  public MonteCarloCounter() {
    setSeed();
    rng = new MRG32k3a();
    entry = new Entry();
  }

  public void setSeed() {
    String[] longs = Config.mcSeed.split(",");
    reseed(longs);
  }

  public void reseed(String[] longs) {
    long[] seeds = new long[6];
    
    for (int i = 0; i < 6; i++) {
      seeds[i] = Long.parseLong(longs[i]);
    }
    MRG32k3a.setPackageSeed(seeds);
  }
  
  public void reset() {
    entry = new Entry();
    domainVolume = 0.0;
  }
  
  public void reset(String seed) {
    reset();
    reseed(seed.split(","));
    rng = new MRG32k3a();
  }
  
  public ResultData.Entry entry;
  public RandomStream rng;
  
  public double domainVolume = 0.0;
  
  @Override
  public EstimateResult count(PC pc) {
    throw new RuntimeException("Not implemented yet");
  }


  /**
   * 'domains' must contain only the intervals for the variables on the PC. 
   */
  
  @Override
  public EstimateResult count(PC pc, Interval[] domains) {
    //use domain as a box
    List<Interval[]> box = new ArrayList<Interval[]>();
    box.add(domains);
    
    double domainVolume = Interval.computeVolume(domains);
    return count(pc,box,domains, domainVolume);
  }
  
//  private static final Interval DUMMY_DOMAIN = new Interval(0, 1);

  /**
   * All interval arrays in 'boxes' must contain only the intervals for 
   * the variables on the PC. 
   */
  
  @Override
  public EstimateResult count(PC pc, List<Interval[]> boxes) {
    //create dummy domain (we already have the boxes from the interval solver)
//    int nVars = boxes.get(0).length;
//    Interval[] domains = new Interval[nVars];
//    for (int i = 0; i < nVars; i++) {
//      domains[i] = DUMMY_DOMAIN;
//    }

    //no domain specified
    double domainVolume = 0;
    for (Interval[] box : boxes) {
      domainVolume += Interval.computeVolume(box);
    }
    
    return count(pc,boxes,new Interval[]{}, domainVolume); 
  }
  
  /**
   * 'domains' must contain only the intervals for the variables on the PC.
   * All interval arrays in 'boxes' must contain only the intervals for 
   * the variables on the PC. 
   */
  
  @Override
  public EstimateResult count(PC pc, List<Interval[]> boxes, Interval[] domain) {
    double domainVolume = Interval.computeVolume(domain);
    return count(pc,boxes,domain,domainVolume);
  }
  
  private EstimateResult count(PC pc, List<Interval[]> boxes, Interval[] domains, double domainVolume) {
    
    this.domainVolume = domainVolume;
    double totalVolEstimate = 0;
    List<EstimateResult> estimates = new ArrayList<EstimateResult>();
    
    int[] samplesPerBox = computeNumSamples(boxes);
    int i = 0;
    int hits = 0;
    
    for (Interval[] box : boxes) {
      Tally stats = new Tally("Monte Carlo tally stats for box " + i);
      stats.init();

      EstimateResult regionResult = simulate(pc, box, stats, samplesPerBox[i]);
      estimates.add(regionResult);
      totalVolEstimate += regionResult.mean;
      hits += stats.sum();

//      System.out.println("Total for box " + i + " : " + partial);
//      System.out.println(stats.report());  
      i++;
    }
    
    //compute variance according to equation (9) on the paper
    double stratifiedVariance = 0;
    
    for (int j = 0; j < estimates.size(); j++) {
      Interval[] box = boxes.get(j);
      double boxWeight = boxWeight(box);
      
      EstimateResult estimate = estimates.get(j);
      double var = estimate.variance; 
      
      stratifiedVariance += Math.pow(boxWeight,2) * var;  
    }
    
    int nBoxes = boxes.size();
    double stratifiedStdev = Math.sqrt(stratifiedVariance);
    
//    System.out.println("TOTAL: " + total);
    nBoxes = nBoxes != 0 ? nBoxes : 1;
    entry.nBoxes = nBoxes;
    entry.stdev = stratifiedStdev;
    entry.estimateResult = totalVolEstimate;
    entry.variance = stratifiedVariance;
    entry.lowerbound95 = 0;
    entry.upperbound95 = 0;
    
//    System.out.printf("[monte carlo] ran for a total of %d iterations (%d boxes, %d iterations for each one)\n",Config.mcMaxSamples, nBoxes, Config.mcMaxSamples / nBoxes);
//    System.out.printf("[monte carlo] hits: %d;\n",hits);
//    System.out.printf("[monte carlo] stratified variance: %f; stdev: %f\n",stratifiedVariance,stratifiedStdev);
//    System.out.printf("[monte carlo] total: %f\n",totalVolEstimate);
    
    //TODO added by Antonio, just for testing
    if (Main.mcResetRngEachSimulation) {
      rng.resetStartStream();
    }
    
    return new EstimateResult(totalVolEstimate, stratifiedVariance);
  }
  
  private int[] computeNumSamples(List<Interval[]> boxes) {
    int[] nsamples = new int[boxes.size()];
    double[] sizes = new double[boxes.size()];
    
    if (Config.mcDistributeSamples) {
      double budget = Config.mcMaxSamples;
      
      double total = 0;
      for (int i = 0; i < boxes.size(); i++) {
        sizes[i] = Interval.computeVolume(boxes.get(i));
      }

      //size which will result in the minimum amount of samples
      double cut = Config.mcMinSamplesPerBox * total / budget;
      double totalMinusCuts = 0;
      for (int i = 0; i < boxes.size(); i++) {
        if (sizes[i] < cut) {
          nsamples[i] = Config.mcMinSamplesPerBox;
          budget = budget - Config.mcMinSamplesPerBox;
        } else {
          totalMinusCuts += sizes[i];
        }
      }
      
      for (int i = 0; i < boxes.size(); i++) {
        if (nsamples[i] == 0) {
          int samples = (int) Math.floor(sizes[i] * budget / totalMinusCuts);
          nsamples[i] = samples;
        }
      }
      
    } else {
      for (int i = 0; i < boxes.size(); i++) {
        nsamples[i] = Config.mcMaxSamples / boxes.size();
      }
    }
    
    int totalSamples = 0;
    for (Integer s : nsamples) {
      totalSamples += s;
    }
//    System.out.println("[sampling] sizes: " + Arrays.toString(sizes));
//    System.out.println("[sampling] boxes: " + toString(boxes));
//    System.out.println("[sampling] domainVolume: " + domainVolume);
//    System.out.println("[sampling] nsamples: " + Arrays.toString(nsamples));
//    System.out.println("[sampling] total: " + totalSamples);
    return nsamples;
  }
  
  private String toString(List<Interval[]> boxes) {
    StringBuffer sb = new StringBuffer();
    
    for (Interval[] box : boxes) {
      sb.append(Arrays.toString(box));
      sb.append(" ## ");
    }
    
    return sb.toString();
  }

  private EstimateResult simulate(PC pc, Interval[] box, Tally stats, int nSamples) {
    double boxWeight = boxWeight(box);
    
    if (boxWeight == 0) {
      System.out.println("[monte carlo] 0-vol box - skipping simulation...");
      return new EstimateResult(0, 0);
    } else {
      Set<SymLiteral> varSet = pc.getSortedVars();
      List<Elem[]> l = prepareRPNState(pc);
      
      for (int i = 0; i < nSamples; i++) {
//      for (SymLiteral lit : varSet) {
//        Interval iv = box[lit.getId() - 1];
//        lit.setCte((rng.nextDouble() * iv.getLength()) + iv.lo);
//      }
        
        Iterator<SymLiteral> it = varSet.iterator();
        for (Interval iv : box) {
          SymLiteral lit = it.next();
          lit.setCte((rng.nextDouble() * iv.getLength()) + iv.lo);
        }
        
        /** reverse polish code **/
        boolean isSAT = true;
        isSAT = evalRNP(l);
        
        stats.add(isSAT ? 1.0 : 0.0);
      }
      
      double avg = stats.average();
      double binomialVariance = (avg * (1 - avg) / stats.numberObs());
      return new EstimateResult(avg * boxWeight, binomialVariance);
    }
  }
  
  /**
   * Returns the weight (relative size) of the box, i.e. volume(box)/domainVolume 
   * @param box
   * @return weight of the box
   */
  
  private double boxWeight(Interval[] box) {
    double weight = 1;

    if (domainVolume == 0) {
      weight = 0;
    } else {
      for (Interval iv : box) {
        if (iv != null) { //null boxes = vars that 
          weight = weight * iv.getLength();
        }
      }
      weight = weight / domainVolume;
    }
    
    return weight;
  }

  private static boolean evalRNP(List<Elem[]> l) {
    boolean result = true;
    for (Elem[] elem : l) {
      ReversePolish rpol = new ReversePolish(elem);
      if (rpol.eval().intValue()==0) {
        result = false;
        break;
      }
    }
    return result;
  }

  private static List<Elem[]> prepareRPNState(PC pc) {
    /** reverse polish code **/
    List<Elem[]> l = new ArrayList<Elem[]>();
    for (SymBool bExp : pc.getConstraints()) {
      GenReversePolishExpression revPolExpr = new GenReversePolishExpression();
      revPolExpr.visitSymBool(bExp);
      l.add(revPolExpr.getElems());  
    }
    return l;
  }

  public static void main(String[] args) throws ParseException, IOException, InterruptedException {
    PC pc = new Parser(Benchmarks.CONE).parsePC();
    runCounter(args[1], pc);
    
//    for (Interval[] box : boxes) {
//      System.out.println(Arrays.toString(box));
//    }
    
/*    String[] benchs = new String[] { Benchmarks.SPHERE, Benchmarks.TETRAHEDRON,
        Benchmarks.CUBE, Benchmarks.ICOSAHEDRON,
        Benchmarks.RHOMBICUBOCTAHEDRON, Benchmarks.SPHERE_SPHERE_INTERSECTION,
        Benchmarks.CONE_CYLINDER };
*/    
//    for (String con : benchs) {
		
//        for (Interval[] box : boxes) {
//          System.out.println(Arrays.toString(box));
//        }
	

//    }
    
//    System.out.println("-------------------------------------------");
//    System.out.println("running without stratified sampling (x10 iterations)");
//    
//    TOTAL_RUNS = TOTAL_RUNS * 10;
//    boxes = isolver.getPaving(pc);
//    boxes = foldBoxes(boxes);
//    for (Interval[] box : boxes) {
//      System.out.println(Arrays.toString(box));
//    }
//    
//    Tally stats2 = new Tally("results for 100 iterations of normal sampling");
//    stats2.init();
//    for (int i = 0; i < 100; i++) {
//      stats2.add(mcc.count(pc, boxes));
//    }
//    System.out.println("-------------------------------------------");
//    System.out.println("-------------------------------------------");
//    System.out.println(stats.report());
//    System.out.println(stats2.report());
//    mcc.count(pc)
  }

  public static ResultData.Entry runCounter(String domainString, PC pc)
      throws IOException, InterruptedException {
    MonteCarloCounter mcc = new MonteCarloCounter();
    IntervalSolverCaller isolver = coral.solvers.rand.Util.getIntervalSolver(Config.intervalSolver);
//    PC pc = new Parser(args[0]).parsePC().getCanonicalForm();
    List<Interval[]> boxes;
    Interval[] domain;
    if (domainString != null) {
      domain = MCUtil.parseDomain(domainString); 
    } else {
      domain = null; //TODO refactor this
    }
    boxes = isolver.getPaving(pc, domain);
    if (boxes.size() == 0) {
      System.out.println("[realpaver] no solutions found for current PC!");
    } else {
      if (domainString != null) {
        mcc.count(pc, boxes, domain);
      } else {
        mcc.count(pc, boxes);
      }
    }
    
    return mcc.entry;
  }

  private static List<Interval[]> foldBoxes(List<Interval[]> boxes) {
    double[] lowest = new double[boxes.get(0).length];
    double[] highest = new double[boxes.get(0).length];
    Arrays.fill(highest, Double.MIN_VALUE);
    Arrays.fill(lowest, Double.MAX_VALUE);
    
    for(Interval[] box : boxes) {
      for (int i=0; i < box.length; i++) {
        if (box[i].lo < lowest[i]) {
          lowest[i] = box[i].lo;
        }
        
        if (box[i].hi > highest[i]) {
          highest[i] = box[i].hi;
        }
      }
    }
    
    Interval[] folded = new Interval[lowest.length];
    for (int i = 0; i < folded.length; i++) {
      folded[i] = new Interval(lowest[i], highest[i]);
    }
    List<Interval[]> list = new ArrayList<Interval[]>(1);
    list.add(folded);
    return list;
  }
}
