package coral.util.callers;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import symlib.SymBool;
import symlib.SymLiteral;

import com.wolfram.jlink.KernelLink;
import com.wolfram.jlink.MathLinkException;
import com.wolfram.jlink.MathLinkFactory;

import coral.PC;
import coral.util.Config;
import coral.util.Interval;
import coral.util.visitors.MathematicaVistor;

public class MathematicaCaller {

  KernelLink ml = null;
  int fCounter = 1;

  public MathematicaCaller() {
    try {
      // Typical launch on Windows
      // KernelLink ml =
      // MathLinkFactory.createKernelLink("-linkmode launch -linkname 'c:\\program files\\wolfram research\\mathematica\\7.0\\mathkernel.exe'");
      // Typical launch on Unix
      ml = MathLinkFactory
          .createKernelLink("-linkmode launch -linkname 'math -mathlink'");

      // Typical launch on Mac OS X
      // ml =
      // MathLinkFactory.createKernelLink("-linkmode launch -linkname '/Applications/Mathematica.app/Contents/MacOS/MathKernel'");

      // Get rid of the initial InputNamePacket the kernel will send
      // when it is launched.
      ml.discardAnswer();
      ml.evaluateToOutputForm("SeedRandom['"+Config.mcSeed+"']", 0);
    } catch (MathLinkException e) {
      System.out.println("Fatal error opening link: " + e.getMessage());
      throw new RuntimeException(e);
    }
  }

  /**
   * The order of the boxes should be the same as the order of the variables
   * returned by PC.getSortedVars()
   * 
   * @param pc
   * @param boxes
   * @return
   */

  public double integrateAdaptativeMonteCarlo(PC pc, List<Interval[]> boxes) {
    List<String> mathCons = new ArrayList<String>();
    MathematicaVistor mv = new MathematicaVistor();

    for (SymBool bool : pc.getConstraints()) {
      mathCons.add(bool.accept(mv));
    }

    Set<SymLiteral> vars = pc.getSortedVars();
    String function = prepareMathematicaFunction(mathCons, vars);
	System.out.println(function);
    String strResult = ml.evaluateToOutputForm(function, 0);
    System.out.println("[mathematica] " + strResult);

    // f[x_,y_] -> f[x,y]
    String functionReference = function.split(":=")[0].replace("_", "");

    double volume = 0.0;
    
    for (Interval[] box : boxes) {
      String command = prepareCommand(box, vars, functionReference);
	  System.out.println(command);
      strResult = ml.evaluateToOutputForm(command, 0);
      System.out.println("[mathematica] " + strResult);
      
      if (strResult.split("E", -1).length - 1 >= 2) { //output is composed by 2 or more numbers
        strResult = strResult.substring(0, strResult.indexOf(' '));
      }
      volume += parseMathematicaDouble(strResult);
      System.out.println(volume);
    }

    return volume;
  }

  private double parseMathematicaDouble(String output) {
    double val; 
    if (output.endsWith(".E")) { //hack to avoid issues with fp representation
      val = Double.parseDouble(output.substring(0, output.length() - 2));
    } else if (output.endsWith("E")) { //hack to avoid issues with fp representation
      val = Double.parseDouble(output.substring(0, output.length() - 1));
    } else {
      val = Double.parseDouble(output);
    }
    return val;
  }

  private String prepareCommand(Interval[] box, Set<SymLiteral> vars,
      String functionReference) {
    StringBuffer sb = new StringBuffer();
    sb.append("ScientificForm[NIntegrate[").append(functionReference);
    
    Iterator<SymLiteral> varIt = vars.iterator(); 
    for(Interval iv : box) {
      String litName = "x" + varIt.next().getId();
      sb.append(", {").append(litName).append(",");
      
      String lo = MathematicaVistor.toMathematicaScientificNotation(iv.lo);
      String hi = MathematicaVistor.toMathematicaScientificNotation(iv.hi);
      
      sb.append(lo).append(", ");
      sb.append(hi).append("}");
    }
    if (Config.mathematicaGA) {
		//    	sb.append(", Method -> {GlobalAdaptive}]");
		sb.append("],NumberFormat -> (#1 <> \"E\" <> #3 &)]");
    } else {
    	sb.append(", Method -> {MonteCarlo, MaxPoints -> ");
    	sb.append(Config.mcMaxSamples);
    	sb.append("}]");
    	sb.append(",NumberFormat -> (#1 <> \"E\" <> #3 &)]");
    }
	//    sb.append(" }, PrecisionGoal -> 5],NumberFormat -> (#1 <> \"E\" <> #3 &)]");
    return sb.toString();
  }

  // ex: f[z_,y_,x_] := If[x>10.5 && x<15.3, 1, 0];
  private String prepareMathematicaFunction(List<String> mathCons,
      Set<SymLiteral> vars) {
    StringBuffer sb = new StringBuffer();
    sb.append("f").append(fCounter++);
    sb.append('[');

    for (SymLiteral var : vars) {
      sb.append("x" + var.getId() + "_,");
    }
    sb.delete(sb.length() - 1, sb.length());
    sb.append("] := If[");

    for (String cons : mathCons) {
      sb.append(cons);
      sb.append(" && ");
    }

    sb.delete(sb.length() - 4, sb.length());
    sb.append(", 1.0, 0.0];");

    return sb.toString();
  }

  public void reseed(String seed) {
    ml.evaluateToOutputForm("SeedRandom['"+seed+"']", 0);
  }
}
